@@ -1,4 +1,4 @@
-from typing import Optional, Callable, Iterator, Iterable, List, Set, Mapping, Tuple, Any
+from typing import Optional, Callable, Iterator, Iterable, List, Set, Mapping, Tuple, Any, Dict
 from functools import cmp_to_key
 import warnings
 import importlib
@@ -98,16 +98,34 @@ class PropertySignature(object):
 
 
 def replace_numpy_array(match_obj):
-    result = r"numpy.ndarray[{type}, _Shape[{shape}]]".format(type=match_obj.group("type"),
-                                                              shape=match_obj.group("shape"))
+    numpy_type = match_obj.group("type")
+    # pybind always append size of data type
+    if numpy_type in ['int8', 'int16', 'int32', 'int64',
+                      'float16', 'float32', 'float64',
+                      'complex32', 'complex64', 'longcomplex'
+                      ]:
+        numpy_type = "numpy." + numpy_type
+
+    shape = match_obj.group("shape")
+    if shape:
+        shape = ", _Shape[{}]".format(shape)
+    else:
+        shape = ""
+    result = r"numpy.ndarray[{type}{shape}]".format(type=numpy_type, shape=shape)
     return result
 
 
+def replace_typing_types(match):
+    # pybind used to have iterator/iterable in place of Iterator/Iterable
+    return "typing." + match.group('type').capitalize()
+
+
 class StubsGenerator(object):
     INDENT = " " * 4
 
     GLOBAL_CLASSNAME_REPLACEMENTS = {
-        re.compile(r"numpy.ndarray\[(?P<type>[^\[\]]+)(\[(?P<shape>[^\[\]]+)\])\]"): replace_numpy_array
+        re.compile(r"numpy.ndarray\[(?P<type>[^\[\]]+)(\[(?P<shape>[^\[\]]+)\])?\]"): replace_numpy_array,
+        re.compile(r"(?<!\w)(?P<type>Callable|Dict|[Ii]terator|[Ii]terable|List|Optional|Set|Tuple|Union')(?!\w)"): replace_typing_types
     }
 
     def parse(self):
@@ -129,7 +147,7 @@ class StubsGenerator(object):
     @staticmethod
     def fully_qualified_name(klass):
         module_name = klass.__module__ if hasattr(klass, '__module__') else None
-        class_name = klass.__qualname__
+        class_name = getattr(klass, "__qualname__", klass.__name__)
 
         if module_name == "builtins":
             return class_name
@@ -333,7 +351,7 @@ class FreeFunctionStubsGenerator(StubsGenerator):
             logger.debug("Docstring is empty for '%s'" % self.fully_qualified_name(self.member))
         for sig in self.signatures:
             if len(self.signatures) > 1:
-                result.append("@overload")
+                result.append("@typing.overload")
             result.append("def {name}({args}) -> {rtype}:".format(
                 name=sig.name,
                 args=sig.args,
@@ -376,7 +394,7 @@ class ClassMemberStubsGenerator(FreeFunctionStubsGenerator):
                 # remove type of self
                 args = ",".join(["self"] + sig.split_arguments()[1:])
             if len(self.signatures) > 1:
-                result.append("@overload")
+                result.append("@typing.overload")
 
             result.append("def {name}({args}) -> {rtype}: {ellipsis}".format(
                 name=sig.name,
@@ -554,6 +572,7 @@ class ModuleStubsGenerator(StubsGenerator):
         self.free_functions = []  # type: List[FreeFunctionStubsGenerator]
         self.submodules = []  # type: List[ModuleStubsGenerator]
         self.imported_modules = []  # type: List[str]
+        self.imported_classes = {}  # type: Dict[str, type]
         self.attributes = []  # type: List[AttributeStubsGenerator]
         self.stub_suffix = ""
         self.write_setup_py = False
@@ -569,16 +588,19 @@ class ModuleStubsGenerator(StubsGenerator):
         for name, member in inspect.getmembers(self.module):
             if inspect.ismodule(member):
                 m = ModuleStubsGenerator(member)
-                if m.module.__name__.startswith(self.module.__name__):
+                if m.module.__name__.split('.')[:-1] == self.module.__name__.split('.'):
                     self.submodules.append(m)
                 else:
-                    self.imported_modules = m.module.__name__
+                    self.imported_modules += [m.module.__name__]
                     logger.debug("Skip '%s' module while parsing '%s' " % (m.module.__name__, self.module.__name__))
             elif inspect.isbuiltin(member) or inspect.isfunction(member):
                 self.free_functions.append(FreeFunctionStubsGenerator(name, member, self.module.__name__))
             elif inspect.isclass(member):
-                if member.__name__ not in self.class_name_blacklist:
-                    self.classes.append(ClassStubsGenerator(member))
+                if member.__module__ == self.module.__name__:
+                    if member.__name__ not in self.class_name_blacklist:
+                        self.classes.append(ClassStubsGenerator(member))
+                else:
+                    self.imported_classes[name] = member
             elif name == "__doc__":
                 self.doc_string = member
             elif name not in self.attributes_blacklist:
@@ -621,7 +643,7 @@ class ModuleStubsGenerator(StubsGenerator):
         for f in self.free_functions:  # type: FreeFunctionStubsGenerator
             result |= f.get_involved_modules_names()
 
-        return set(result) - {"builtins", self.module.__name__}
+        return set(result) - {"builtins", 'typing', self.module.__name__}
 
     def to_lines(self):  # type: () -> List[str]
 
@@ -636,18 +658,18 @@ class ModuleStubsGenerator(StubsGenerator):
 
         # import everything from typing
         result += [
-            "from typing import *"
-        ]
-        # pybind11 has some lower-cased prototypes from typing
-        result += [
-            "from typing import Iterable as iterable",
-            "from typing import Iterator as iterator",
+            "import typing"
         ]
 
-        result += [
-            "from numpy import float64",
-            "_Shape = Tuple[int, ...]"
-        ]
+        for name, class_ in self.imported_classes.items():
+            class_name = getattr(class_, "__qualname__", class_.__name__)
+            if name == class_name:
+                suffix = ""
+            else:
+                suffix = " as {}".format(name)
+            result += [
+                'from {} import {}{}'.format(class_.__module__, class_name, suffix)
+            ]
 
         # import used packages
         used_modules = sorted(self.get_involved_modules_names())
@@ -656,9 +678,17 @@ class ModuleStubsGenerator(StubsGenerator):
             # result.extend(map(self.indent, map(lambda m: "import {}".format(m), used_modules)))
             result.extend(map(lambda mod: "import {}".format(mod), used_modules))
 
+        if "numpy" in used_modules:
+            result += [
+                "_Shape = typing.Tuple[int, ...]"
+            ]
+
         # define __all__
 
         _all_ = []
+        for name in self.imported_classes.keys():
+            _all_.append(name)
+
         for c in self.classes:
             _all_.append(c.klass.__name__)
 
